from flask import Flask


def create_app(config):
    app = Flask('tamarack')

    configure_app(app, config)
    configure_blueprints(app)
    configure_db(app)
    configure_celery(app)

    return app


def configure_app(app, config):
    app.config.from_pyfile('settings/defaults.py')
    app.config.from_pyfile(config)


def configure_blueprints(app):
    from tamarack.receiverapi.views import receiverapi
    from tamarack.explorerapi.views import explorerapi
    from tamarack.explorerweb.views import explorerweb

    app.register_blueprint(receiverapi, url_prefix='/receiver-api/v1')
    app.register_blueprint(explorerapi, url_prefix='/explorer-api/v1')
    app.register_blueprint(explorerweb, url_prefix='/')


def configure_db(app):
    from .db import db

    db.init_app(app)


def configure_celery(app):
    from .celery import celery

    celery.conf.update(app.config)

    TaskBase = celery.Task

    class ContextTask(TaskBase):
        abstract = True

        def __call__(self, *args, **kwargs):
            with app.app_context():
                return super().__call__(*args, **kwargs)

    celery.Task = ContextTask



import contextlib

from flask.ext.sqlalchemy import SQLAlchemy

db = SQLAlchemy()


@contextlib.contextmanager
def locked_table(model, lockmode='EXCLUSIVE'):
    lockmodes = ['EXCLUSIVE']
    lockmode = lockmode.upper()

    assert lockmode in lockmodes

    db.session.begin_nested()
    db.session.execute('LOCK TABLE %s IN %s MODE'
                       % (model.__tablename__, lockmode))
    try:
        yield
        db.session.commit()
    except:
        db.session.rollback()
        raise



USEC_TO_MSEC = 1 / 1000


def ms_per_req(datapoint):
    request_count = datapoint.request_count

    def val_per_req(x):
        return (USEC_TO_MSEC * x / request_count) if request_count else None

    return {k: val_per_req(v) for k, v in datapoint.sensor_data.items()}


def reqs_per_min(datapoint):
    return {'request_count': datapoint.request_count}


def errs_per_req(datapoint):
    request_count = datapoint.request_count
    error_count = datapoint.error_count

    return {'error_count': (error_count / request_count) if request_count else None}


def total_time(datapoint):
    return {'total_time': sum(datapoint.sensor_data.values()) * USEC_TO_MSEC}


MAPPERS = {
    'ms-per-req': ms_per_req,
    'reqs-per-min': reqs_per_min,
    'errs-per-req': errs_per_req,
    'total-time': total_time,
}


def mapper_with_name(name):
    return MAPPERS[name]



from flask import abort

from .edn import ednify
from .generic import data_response


def get_one_or_404(query):
    obj = query.first()

    if not obj:
        abort(404)

    return obj


__all__ = ['ednify', 'data_response', 'get_one_or_404']



"""update_counters function

Revision ID: 4026a7a9d9f
Revises: 23e475d6af9
Create Date: 2014-07-12 18:07:14.843760

"""

# revision identifiers, used by Alembic.
revision = '4026a7a9d9f'
down_revision = '23e475d6af9'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.execute('''
        CREATE OR REPLACE FUNCTION hstore_sum
            (lhs hstore, rhs hstore)
            RETURNS hstore
            LANGUAGE sql
            IMMUTABLE

        AS $function$
            SELECT hstore(array_agg(key), array_agg(value::text))
                FROM (SELECT key, SUM(value::bigint) AS value
                    FROM (
                        SELECT (each(lhs)).*
                        UNION ALL
                        SELECT (each(rhs)).*
                    ) AS data
                    GROUP BY key
                ) AS agg
        $function$
    ''')

    op.execute('''
        CREATE OR REPLACE FUNCTION update_request_by_minute
            (in_app_id integer, in_timestamp timestamp,
                sensor_data_delta hstore, request_count_delta bigint, error_count_delta bigint)
            RETURNS VOID
            LANGUAGE plpgsql
    AS $function$
        BEGIN
            LOOP
                UPDATE request_by_minute SET
                    sensor_data = hstore_sum(sensor_data, sensor_data_delta),
                    request_count = request_count + request_count_delta,
                    error_count = error_count + error_count_delta
                WHERE
                    app_id = in_app_id AND
                    timestamp = in_timestamp;

                IF found THEN
                    RETURN;
                END IF;

                BEGIN
                    INSERT INTO request_by_minute
                        (app_id, timestamp, sensor_data, request_count, error_count)
                        VALUES (
                            in_app_id,
                            in_timestamp,
                            sensor_data_delta,
                            request_count_delta,
                            error_count_delta
                        );
                    RETURN;
                EXCEPTION WHEN unique_violation THEN
                    -- continue loop
                END;
            END LOOP;
        END;
    $function$
    ''')

    op.execute('''
        CREATE OR REPLACE FUNCTION update_request_endpoint_by_minute
            (in_app_id integer, in_timestamp timestamp, in_endpoint varchar(100),
                sensor_data_delta hstore, request_count_delta bigint, error_count_delta bigint)
            RETURNS VOID
            LANGUAGE plpgsql
    AS $function$
        BEGIN
            LOOP
                UPDATE request_endpoint_by_minute SET
                    sensor_data = hstore_sum(sensor_data, sensor_data_delta),
                    request_count = request_count + request_count_delta,
                    error_count = error_count + error_count_delta
                WHERE
                    app_id = in_app_id AND
                    timestamp = in_timestamp AND
                    endpoint = in_endpoint;

                IF found THEN
                    RETURN;
                END IF;

                BEGIN
                    INSERT INTO request_endpoint_by_minute
                        (app_id, timestamp, endpoint, sensor_data, request_count, error_count)
                        VALUES (
                            in_app_id,
                            in_timestamp,
                            in_endpoint,
                            sensor_data_delta,
                            request_count_delta,
                            error_count_delta
                        );
                    RETURN;
                EXCEPTION WHEN unique_violation THEN
                    -- continue loop
                END;
            END LOOP;
        END;
    $function$
    ''')


def downgrade():
    op.execute('''
        DROP FUNCTION IF EXISTS hstore_sum(hstore, hstore)
    ''')
    op.execute('''
        DROP FUNCTION IF EXISTS update_request_by_minute
            (integer, timestamp, hstore, bigint, bigint)
    ''')
    op.execute('''
        DROP FUNCTION IF EXISTS update_request_endpoint_by_minute
            (integer, timestamp, hstore, bigint, bigint)
    ''')



"""incr_key function

Revision ID: 23e475d6af9
Revises: 5a3a03e932e
Create Date: 2014-06-29 12:47:45.479186

"""

# revision identifiers, used by Alembic.
revision = '23e475d6af9'
down_revision = '5a3a03e932e'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.execute('''
    CREATE OR REPLACE FUNCTION incr_key(field hstore, key text, val bigint)
        RETURNS hstore
        LANGUAGE sql
        IMMUTABLE
    AS $function$
        SELECT field || HSTORE(key, COALESCE((field -> key)::BIGINT + val, val)::TEXT)
    $function$
    ''')


def downgrade():
    op.execute('DROP FUNCTION incr_key(hstore, text, bigint)')



"""hstore extension

Revision ID: 4989fdc8401
Revises: 23e475d6af9
Create Date: 2014-06-29 13:58:45.148809

"""

# revision identifiers, used by Alembic.
revision = '4989fdc8401'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.execute('CREATE EXTENSION IF NOT EXISTS hstore')


def downgrade():
    op.execute('DROP EXTENSION hstore')



from ..db import db


class Application(db.Model):
    __tablename__ = 'application'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(80), nullable=False, unique=True)
    display_name = db.Column(db.String(80), nullable=False)

    def __init__(self, name):
        self.name = name
        self.display_name = name

    def as_data(self):
        return {
            'id': self.id,
            'name': self.name,
            'display_name': self.display_name
        }

    @classmethod
    def by_name(cls, name):
        app = Application.query.filter(cls.name == name).first()

        if not app:
            app = Application(name)
            db.session.add(app)

        return app



from sqlalchemy.dialects.postgresql import HSTORE

from ..db import db, locked_table


class RequestByMinute(db.Model):
    __tablename__ = 'request_by_minute'

    app_id = db.Column(db.Integer, db.ForeignKey('application.id'), primary_key=True)
    timestamp = db.Column(db.DateTime(timezone=True), primary_key=True)
    sensor_data = db.Column(HSTORE, nullable=False)
    request_count = db.Column(db.BigInteger, nullable=False)
    error_count = db.Column(db.BigInteger, nullable=False)

    app = db.relationship('Application',
                          backref=db.backref('request_by_minute', lazy='dynamic'))

    def __init__(self, app, timestamp):
        self.app = app
        self.timestamp = timestamp
        self.sensor_data = {}
        self.request_count = 0
        self.error_count = 0


class RequestEndpointByMinute(db.Model):
    __tablename__ = 'request_endpoint_by_minute'

    app_id = db.Column(db.Integer, db.ForeignKey('application.id'), primary_key=True)
    endpoint = db.Column(db.String(100), primary_key=True)
    timestamp = db.Column(db.DateTime(timezone=True), primary_key=True)
    sensor_data = db.Column(HSTORE, nullable=False)
    request_count = db.Column(db.BigInteger, nullable=False)
    error_count = db.Column(db.BigInteger, nullable=False)

    app = db.relationship('Application',
                          backref=db.backref('request_endpoint_by_minute', lazy='dynamic'))

    def __init__(self, app, timestamp, endpoint):
        self.app = app
        self.timestamp = timestamp
        self.endpoint = endpoint
        self.sensor_data = {}
        self.request_count = 0
        self.error_count = 0


def process_minute_datapoint(app, datapoint):
    timestamp = datapoint['timestamp']
    request_count = datapoint['request_count']
    error_count = datapoint['error_count']
    endpoint = datapoint['endpoint']
    sensor_data = {k: str(v) for k, v in datapoint['sensor_data'].items()}

    db.session.execute(
        db.func.update_request_by_minute(
            app.id, timestamp,
            sensor_data, request_count, error_count))

    db.session.execute(
        db.func.update_request_endpoint_by_minute(
            app.id, timestamp, endpoint,
            sensor_data, request_count, error_count))


@db.event.listens_for(RequestByMinute, 'load')
@db.event.listens_for(RequestEndpointByMinute, 'load')
def make_int_sensor_values(target, context):
    target.sensor_data = {k: int(v) for k, v in target.sensor_data.items()}



from edn_format import Keyword, dumps
from flask import current_app


def keywordize(s):
    return Keyword(s.replace('_', '-'))


def py_to_clj(v):
    if isinstance(v, dict):
        return {(keywordize(k) if isinstance(k, str) else k): py_to_clj(v)
                for k, v in v.items()}

    if any(isinstance(v, c) for c in (list, tuple, set, frozenset)):
        return v.__class__([py_to_clj(x) for x in v])

    return v


def ednify(data):
    return current_app.response_class(
        dumps(py_to_clj(data)).encode('utf-8'),
        mimetype='application/edn; charset=UTF-8')



import click
import os
import shutil

from alembic import command
from alembic.config import Config
from werkzeug import script

from tamarack.core.app import create_app


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.option('--config', metavar='PYTHONCLASS',
              default='tamarack.settings.development.Config',
              help='Python class used as configuration')
@click.pass_context
def cli(ctx, config):
    '''Tamarack manager and runner'''

    ctx.obj = {
        'config': os.path.join(os.getcwd(), os.path.expanduser(config)) if config else None
    }


@cli.command()
@click.option('--enable-profiler', default=False, is_flag=True,
              help='Start the server with a Python profiler on each request')
@click.pass_context
def devserver(ctx, enable_profiler):
    '''Starts a development server'''

    app = create_app(ctx.obj['config'])

    if enable_profiler:
        from werkzeug.contrib.profiler import ProfilerMiddleware
        app.wsgi_app = ProfilerMiddleware(app.wsgi_app, restrictions=[30])

    app.run(debug=app.config['DEBUG'], port=app.config['PORT'], host=app.config['LISTEN_ADDRESS'])


@cli.command()
@click.pass_context
def server(ctx):
    '''Starts a production HTTP server'''
    import asyncio

    from aiohttp.wsgi import WSGIServerHttpProtocol

    app = create_app(ctx.obj['config'])

    loop = asyncio.get_event_loop()

    def create_server():
        return WSGIServerHttpProtocol(app, readpayload=True)

    future = loop.create_server(create_server,
                                app.config['LISTEN_ADDRESS'],
                                app.config['PORT'])

    loop.run_until_complete(future)

    try:
        print('Tamarack server started')
        loop.run_forever()
    except KeyboardInterrupt:
        pass


@cli.command()
@click.pass_context
def upgrade(ctx):
    '''Upgrade the Tamarack installation'''

    ctx.forward(db)
    ctx.forward(db_upgrade, revision='head', sql=False)


@cli.command()
@click.pass_context
def shell(ctx):
    '''Start a Tamarack-enabled development shell'''

    def make_shell_env():
        from tamarack.core import models
        from tamarack.core.db import db

        app = create_app(ctx.obj['config'])
        request = app.test_request_context('/')
        request.push()

        return {
            'app': app,
            'models': models,
            'request': request,
            'db': db,
        }

    script.make_shell(make_shell_env,
                      banner='Interactive Tamarack Shell')()


@cli.command()
@click.argument('config')
def init(config):
    '''Generate a Tamarack configuration file'''

    srcpath = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                           '../templates/settings.py')
    destpath = os.path.join(os.getcwd(), os.path.expanduser(config))

    shutil.copy(srcpath, destpath)

    click.echo('Tamarack settings file created at %s' % config)


@cli.command()
@click.pass_context
def worker(ctx):
    app = create_app(ctx.obj['config'])

    from tamarack.core.celery import celery

    return celery.start()


@cli.group()
@click.pass_context
def db(ctx):
    '''Perform database operations'''

    app = create_app(ctx.obj['config'])

    file_dir = os.path.abspath(os.path.dirname(__file__))

    config = Config(os.path.join(file_dir, '../alembic.ini'))
    config.set_section_option('alembic', 'script_location',
                              os.path.join(file_dir, '../migrations'))
    config.set_section_option('alembic', 'sqlalchemy.url',
                              app.config['SQLALCHEMY_DATABASE_URI'])

    ctx.obj['db_cfg'] = config


@db.command('revision')
@click.option('--message', '-m', default=None,
              help="Message string to use with 'revision'")
@click.option('--sql', default=False, is_flag=True,
              help="Don't emit SQL to database - dump to standard output/file instead")
@click.option('--autogenerate', default=False, is_flag=True,
              help="Populate revision script with candidate migration operations, based on comparison of database to model.")
@click.pass_context
def db_revision(ctx, message, autogenerate, sql):
    '''Create a new revision file'''

    command.revision(ctx.obj['db_cfg'],
                     message=message,
                     autogenerate=autogenerate,
                     sql=sql)


@db.command('upgrade')
@click.argument('revision')
@click.option('--sql', default=False, is_flag=True,
              help="Don't emit SQL to database - dump to standard output/file instead")
@click.pass_context
def db_upgrade(ctx, revision, sql):
    '''Upgrade to a later version'''

    command.upgrade(ctx.obj['db_cfg'],
                    revision=revision,
                    sql=sql)


@db.command('downgrade')
@click.argument('revision')
@click.option('--sql', default=False, is_flag=True,
              help="Don't emit SQL to database - dump to standard output/file instead")
@click.pass_context
def db_downgrade(ctx, revision, sql):
    '''Revert to a previous version'''

    command.downgrade(ctx.obj['db_cfg'],
                      revision=revision,
                      sql=sql)


@db.command('history')
@click.option('--rev-range', '-r', default=None,
              help="Specify a revision range; format is [start]:[end]")
@click.pass_context
def db_history(ctx, rev_range):
    '''List changesets in chronological order'''

    command.history(ctx.obj['db_cfg'],
                    rev_range=rev_range)



"""Initial revision

Revision ID: 5a3a03e932e
Revises: None
Create Date: 2014-06-27 21:55:35.102633

"""

# revision identifiers, used by Alembic.
revision = '5a3a03e932e'
down_revision = '4989fdc8401'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

def upgrade():
    op.create_table('application',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=80), nullable=False),
    sa.Column('display_name', sa.String(length=80), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name')
    )
    op.create_table('request_endpoint_by_minute',
    sa.Column('app_id', sa.Integer(), nullable=False),
    sa.Column('endpoint', sa.String(length=100), nullable=False),
    sa.Column('timestamp', sa.DateTime(timezone=True), nullable=False),
    sa.Column('sensor_data', postgresql.HSTORE(), nullable=False),
    sa.Column('request_count', sa.BigInteger(), nullable=False),
    sa.Column('error_count', sa.BigInteger(), nullable=False),
    sa.ForeignKeyConstraint(['app_id'], ['application.id'], ),
    sa.PrimaryKeyConstraint('app_id', 'endpoint', 'timestamp')
    )
    op.create_table('request_by_minute',
    sa.Column('app_id', sa.Integer(), nullable=False),
    sa.Column('timestamp', sa.DateTime(timezone=True), nullable=False),
    sa.Column('sensor_data', postgresql.HSTORE(), nullable=False),
    sa.Column('request_count', sa.BigInteger(), nullable=False),
    sa.Column('error_count', sa.BigInteger(), nullable=False),
    sa.ForeignKeyConstraint(['app_id'], ['application.id'], ),
    sa.PrimaryKeyConstraint('app_id', 'timestamp')
    )


def downgrade():
    op.drop_table('request_by_minute')
    op.drop_table('request_endpoint_by_minute')
    op.drop_table('application')



from tamarack.core.db import db
from tamarack.core.celery import celery
from tamarack.core.models import Application
from tamarack.core.models.request import process_minute_datapoint


@celery.task()
def process_request_data(data):
    app = Application.by_name(data['app_name'])

    for minute_data in data['by_minute']:
        process_minute_datapoint(app, minute_data)

    db.session.commit()



import shutil
import subprocess
import sys

from distutils.command.sdist import sdist as SdistCommand
from os import path
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from pathlib import Path

thisdir = Path(__file__).parent.resolve()
thisdir = thisdir


class TamarackSdistCommand(SdistCommand):
    def run_lein(self):
        print('running lein cljsbuild in frontend dir')

        result = subprocess.call('lein cljsbuild once tamarack tamarack-release',
                                 shell=True,
                                 cwd=bytes(thisdir))

        if result:
            raise Exception('Could not compile ClojureScript')

    def run(self):
        self.run_lein()

        super().run()


class PyTestCommand(TestCommand):
    def run_tests(self):
        import pytest
        sys.exit(pytest.main())


tests_requires = [
    'pytest>=2.5,<2.6',
    'flake8>=2.0.0,<3.0.0',
]

redis_requires = [
    'redis>=2.10.0,<2.11.0',
]


setup(
    name='tamarack',
    version='0.0.4',
    description='A quantitative web application profiler',
    url='https://github.com/mhallin/tamarack',

    author='Magnus Hallin',
    author_email='mhallin@gmail.com',

    license='BSD',

    packages=find_packages(exclude=['tests']),

    include_package_data=True,
    zip_safe=False,

    install_requires=[
        'flask>=0.10.0,<0.11.0',
        'click>=2.0,<3.0',
        'SQLAlchemy>=0.9.6,<1.0.0',
        'alembic>=0.6.0,<0.7.0',
        'Flask-SQLAlchemy>=1.0,<2.0',
        'edn_format>=0.5,<0.6',
        'psycopg2>=2.5,<2.6',
        'aiohttp',
        'celery>=3.1.0,<3.2.0',
    ],

    extras_require={
        'tests': tests_requires,
        'redis': redis_requires,
    },

    entry_points={
        'console_scripts': [
            'tamarack = tamarack.runner.main:cli'
        ],
    },

    cmdclass={
        'sdist': TamarackSdistCommand,
        'test': PyTestCommand,
    }
)



from collections import OrderedDict
from flask import jsonify, request

from .edn import ednify

GENERIC_MIME_TYPES = OrderedDict([
    ('application/json',
     lambda d: jsonify(**d) if isinstance(d, dict) else jsonify(items=d)),
    ('application/edn', ednify),
])

DEFAULT_MIME_TYPE = 'application/json'


def data_response(data):
    best_mime = request.accept_mimetypes.best_match(GENERIC_MIME_TYPES.keys())
    conversion = GENERIC_MIME_TYPES.get(best_mime)

    if not conversion:
        conversion = GENERIC_MIME_TYPES[DEFAULT_MIME_TYPE]

    return conversion(data)



from flask import Blueprint, request

from tamarack.common import data_response
from tamarack.core.db import db
from tamarack.core.models import Application
from tamarack.core.models.request import process_minute_datapoint
from tamarack.tasks.request import process_request_data

receiverapi = Blueprint('receiverapi', __name__)


@receiverapi.route('/request-data', methods=['POST'])
def handle_request_data():
    data = request.get_json()
    process_request_data.delay(data)

    return data_response({'status': 'ok'})



import operator

from collections import defaultdict
from flask import Blueprint, request

from tamarack.common import data_response, get_one_or_404
from tamarack.core.models import Application, RequestByMinute, RequestEndpointByMinute
from tamarack.core.mappers import mapper_with_name

explorerapi = Blueprint('explorerapi', __name__)


@explorerapi.route('/applications')
def application_list():
    apps = Application.query.order_by('display_name')
    return data_response([app.as_data() for app in apps])


@explorerapi.route('/applications/<app_name>')
def application_info(app_name):
    app = get_one_or_404(Application.query.filter_by(name=app_name))

    return data_response(app.as_data())


@explorerapi.route('/applications/<app_name>/chart/<chart_type>')
def application_chart(app_name, chart_type):
    app = get_one_or_404(Application.query.filter_by(name=app_name))

    from_timestamp = request.args['from']
    to_timestamp = request.args['to']
    mapper = mapper_with_name(chart_type)

    datapoints = app.request_by_minute.filter(
        RequestByMinute.timestamp >= from_timestamp,
        RequestByMinute.timestamp <= to_timestamp)

    return data_response({
        int(datapoint.timestamp.timestamp()) * 1000: mapper(datapoint) for datapoint in datapoints
    })


def aggregate_endpoint_data(mapper, datapoints):
    endpoint_aggregates = defaultdict(lambda: {'count': 0, 'sum': 0})

    for datapoint in datapoints:
        aggregate = endpoint_aggregates[datapoint.endpoint]
        aggregate['sum'] += sum(mapper(datapoint).values())
        aggregate['count'] += 1

    averaged = {endpoint: aggregate['sum'] / aggregate['count']
                for endpoint, aggregate in endpoint_aggregates.items()}

    return averaged


@explorerapi.route('/applications/<app_name>/aggregate/<agg_type>')
def application_aggregate(app_name, agg_type):
    app = get_one_or_404(Application.query.filter_by(name=app_name))

    from_timestamp = request.args['from']
    to_timestamp = request.args['to']
    mapper = mapper_with_name(agg_type)

    datapoints = list(app.request_endpoint_by_minute.filter(
        RequestEndpointByMinute.timestamp >= from_timestamp,
        RequestEndpointByMinute.timestamp <= to_timestamp))

    averaged_aggregates = aggregate_endpoint_data(mapper, datapoints)

    sorted_aggregates = sorted(
        [(endpoint, value) for endpoint, value in averaged_aggregates.items()],
        key=operator.itemgetter(1),
        reverse=True)

    return data_response(sorted_aggregates)


@explorerapi.route('/applications/<app_name>/endpoints/<endpoint>/chart/<chart_type>')
def application_endpoint_chart(app_name, endpoint, chart_type):
    app = get_one_or_404(Application.query.filter_by(name=app_name))

    from_timestamp = request.args['from']
    to_timestamp = request.args['to']
    mapper = mapper_with_name(chart_type)

    datapoints = app.request_endpoint_by_minute.filter(
        RequestEndpointByMinute.timestamp >= from_timestamp,
        RequestEndpointByMinute.timestamp <= to_timestamp)

    return data_response({
        int(datapoint.timestamp.timestamp()) * 1000: mapper(datapoint) for datapoint in datapoints
    })
